from typing import Hashable, Iterable, List

from .._internal import API


@API.public
class AntidoteError(Exception):
    """ Base class of all errors of antidote. """

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"


@API.public
class DoubleInjectionError(AntidoteError):
    """
    Raised when injecting a function/method that already has been injected.
    """

    def __init__(self, func: object) -> None:
        super().__init__(f"Object {func} has already been injected by Antidote.")


@API.public
class DuplicateDependencyError(AntidoteError):
    """
    A dependency already exists with the same id.
    *May* be raised by _providers.
    """


@API.public
class DependencyInstantiationError(AntidoteError):
    """
    The dependency could not be instantiated.
    """

    def __init__(self, dependency: Hashable, stack: List[Hashable] = None) -> None:
        from .._internal.utils import debug_repr
        msg = f"Could not instantiate {debug_repr(dependency)}"
        stack = (stack or [])
        if stack:  # first and last dependency will raise their own errors.
            stack.append(dependency)
            msg += f"\nFull dependency stack:\n{_stack_repr(stack)}\n"

        super().__init__(msg)


@API.public
class DependencyCycleError(AntidoteError):
    """
    A dependency cycle is found.
    """

    def __init__(self, stack: List[Hashable]) -> None:
        super().__init__(f"Cycle:\n{_stack_repr(stack)}\n")


@API.public
class DependencyNotFoundError(AntidoteError):
    """
    The dependency could not be found.
    """

    def __init__(self, dependency: Hashable) -> None:
        from .._internal.utils import debug_repr
        super().__init__(debug_repr(dependency))


@API.public
class FrozenWorldError(AntidoteError):
    """
    An action failed because the world is frozen. Typically happens when trying
    to register a dependency after having called freeze() on the world.
    """


@API.private
class DebugNotAvailableError(AntidoteError):
    """
    Currently provider do not have to implement the debug behavior. If not, this error
    will be raised and discarded (a warning may be emitted).
    """


@API.private
def _stack_repr(stack: Iterable[object]) -> str:
    from .._internal.utils import debug_repr
    import textwrap

    text = []
    for depth, dependency in enumerate(stack):
        indent = '    ' * (depth - 1) if depth > 1 else ''
        first_line, *rest = debug_repr(dependency).split("\n", 1)
        text.append(f"{indent}{'└── ' if depth else ''}{first_line}")
        if rest:
            text.append(textwrap.indent(rest[0], indent + ('    ' if depth > 1 else '')))
    return '\n'.join(text)



import inspect
from abc import ABCMeta
from typing import Dict, Tuple, Type, cast

from ._internal import API
from ._internal.utils import AbstractMeta
from ._providers import ServiceProvider
from ._providers.service import Parameterized
from ._utils import validate_method_parameters
from .core import Provide, inject

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class ServiceMeta(AbstractMeta):
    def __new__(mcs: 'Type[ServiceMeta]',
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                **kwargs: object
                ) -> 'ServiceMeta':
        cls = cast(
            ServiceMeta,
            super().__new__(mcs, name, bases, namespace, **kwargs)  # type: ignore
        )
        if not kwargs.get('abstract'):
            _configure_service(cls)
        return cls

    @API.public
    def parameterized(cls, **kwargs: object) -> object:
        """
        Creates a new dependency based on the service with the given arguments. The new
        dependency will have the same scope as the original one.

        The recommended usage is to provide a classmethod exposing only parameters that
        may be changed:

        .. doctest:: service_meta

            >>> from antidote import Service, world
            >>> class Database(Service):
            ...     __antidote__ = Service.Conf(parameters=['host'])
            ...
            ...     def __init__(self, host: str):
            ...         self.host = host
            ...
            ...     @classmethod
            ...     def with_host(cls, host: str) -> object:
            ...         return cls.parameterized(host=host)
            >>> db = world.get(Database.with_host(host='remote'))
            >>> db.host
            'remote'
            >>> # As Database is defined as a singleton, the same is applied:
            ... world.get(Database.with_host(host='remote')) is db
            True

        Args:
            **kwargs: Arguments passed on to :code:`__init__()`.

        Returns:
            Dependency to be retrieved from Antidote. You cannot use it directly.
        """
        from .service import Service
        # Guaranteed through _configure_service()
        conf = cast(Service.Conf, getattr(cls, '__antidote__'))
        if conf.parameters is None:
            raise RuntimeError(f"Service {cls} does not accept any parameters. You must "
                               f"specify them explicitly in the configuration with: "
                               f"Service.Conf(parameters=...))")

        if set(kwargs.keys()) != conf.parameters:
            raise ValueError(f"Given parameters do not match expected ones. "
                             f"Got: ({','.join(map(repr, kwargs.keys()))}) "
                             f"Expected: ({','.join(map(repr, conf.parameters))})")

        return Parameterized(cls, kwargs)


@API.private
class ABCServiceMeta(ServiceMeta, ABCMeta):
    pass


@API.private
@inject
def _configure_service(cls: type,
                       service_provider: Provide[ServiceProvider] = None,
                       conf: object = None) -> None:
    from .service import Service
    assert service_provider is not None

    conf = conf or getattr(cls, '__antidote__', None)
    if not isinstance(conf, Service.Conf):
        raise TypeError(f"Service configuration (__antidote__) is expected to be a "
                        f"{Service.Conf}, not a {type(conf)}")

    wiring = conf.wiring

    if wiring is not None:
        wiring.wire(cls)

    validate_method_parameters(cls.__init__, conf.parameters)  # type: ignore

    service_provider.register(cls, scope=conf.scope)



import inspect
from typing import (Callable, FrozenSet, Iterable, Optional, TypeVar, Union, cast,
                    overload)

from ._compatibility.typing import Protocol, final, get_type_hints
from ._factory import FactoryMeta, FactoryWrapper
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from ._internal.wrapper import is_wrapper
from ._providers import FactoryProvider
from ._utils import validated_parameters
from .core import Provide, Scope, Wiring, WithWiringMixin, inject
from .core.exceptions import DoubleInjectionError
from .utils import validated_scope

F = TypeVar('F', bound=Callable[..., object])


@API.private
class FactoryProtocol(Protocol[F]):
    """
    :meta private:
    """

    def __rmatmul__(self, klass: type) -> object:
        pass  # pragma: no cover

    __call__: F


@API.public
class Factory(metaclass=FactoryMeta, abstract=True):
    """
    Defines sublcass as a factory to Antidote. The provided dependency is defined through
    the type annotation of :py:meth:`.__call__`:

    .. doctest:: factory_class

        >>> from antidote import Factory
        >>> class Database:
        ...     pass
        ...
        >>> class DatabaseLoader(Factory):
        ...     def __call__(self) -> Database:
        ...         return Database()

    To retrieve the dependency from Antidote you need to use a specific syntax
    :code:`dependency @ factory` as presented in the following examples. The goal of it is
    twofold:

    - Ensure that the factory is loaded whenever you require the dependency.
    - Better maintainability as you know *where* the dependency comes from.

    .. doctest:: factory_class

        >>> from antidote import world, inject
        >>> world.get(Database @ DatabaseLoader)  # treated as `object` by Mypy
        <Database ...>
        >>> # With Mypy casting
        ... world.get[Database](Database @ DatabaseLoader)
        <Database ...>
        >>> # Concise Mypy casting
        ... world.get[Database] @ DatabaseLoader
        <Database ...>
        >>> @inject([Database @ DatabaseLoader])
        ... def f(db: Database):
        ...     pass

    Or with annotated type hints:

    .. doctest:: factory_class

        >>> from typing import Annotated
        ... # from typing_extensions import Annotated # Python < 3.9
        >>> from antidote import From
        >>> @inject
        ... def f(db: Annotated[Database, From(DatabaseLoader)]):
        ...     pass

    .. note::

        If you only need a simple function, consider using :py:func:`.factory` instead. It
        behaves the same way as above

    All methods are injected by default and the factory returns a singleton by default.
    All of this can be configured with :py:attr:`.__antidote__`:

    .. doctest:: factory_class

        >>> # Singleton by default
        ... world.get[Database] @ DatabaseLoader is world.get[Database] @ DatabaseLoader
        True
        >>> class Session:
        ...     pass
        >>> # The factory will be called anew each time a `Session` is needed. But the
        ... # factory itself, `SessionFactory`, will only be created once.
        ... class SessionFactory(Factory):
        ...     __antidote__ = Factory.Conf(singleton=False)
        ...
        ...     def __init__(self, db: Annotated[Database, From(DatabaseLoader)]):
        ...         self.db = db
        ...
        ...     def __call__(self) -> Session:
        ...         return Session()
        ...
        >>> world.get[Session] @ SessionFactory is world.get[Session] @ SessionFactory
        False

    You may also parameterize the service. Parameters will be passed on to
    :code:`__call__()`:

    .. doctest:: factory_class

        >>> class Database:
        ...     def __init__(self, host: str):
        ...         self.host = host
        ...
        >>> class DatabaseFactory(Factory):
        ...     __antidote__ = Factory.Conf(parameters=['host'])
        ...
        ...     def __call__(self, host: str) -> Database:
        ...         return Database(host)
        ...
        ...     @classmethod
        ...     def hosted(cls, host: str) -> object:
        ...          return cls.parameterized(host=host)
        ...
        >>> test_db = world.get[Database] @ DatabaseFactory.hosted('test')
        >>> test_db.host
        'test'
        >>> # The factory returns a singleton so our test_session will also be one
        ... world.get[Database] @ DatabaseFactory.hosted('test') is test_db
        True

    """

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable factory configuration. To change parameters on a existing instance, use
        either method :py:meth:`.copy` or
        :py:meth:`.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'scope', 'parameters')
        wiring: Optional[Wiring]
        scope: Optional[Scope]
        parameters: Optional[FrozenSet[str]]

        @property
        def singleton(self) -> bool:
            return self.scope is Scope.singleton()

        def __init__(self,
                     *,
                     wiring: Optional[Wiring] = Wiring(),
                     singleton: bool = None,
                     scope: Optional[Scope] = Scope.sentinel(),
                     parameters: Iterable[str] = None):
            """

            Args:
                wiring: Wiring to be applied on the factory. By default only
                    :code:`__init__()` and :code:`__call__()` will be wired. To deactivate
                    any wiring at all use :py:obj:`None`.
                singleton: Whether the returned dependency  is a singleton or not. If yes,
                    the factory will be called at most once and the result re-used.
                    Mutually exclusive with :code:`scope`. Defaults to :py:obj:`True`
                scope: Scope of the returned dependency. Mutually exclusive with
                    :code:`singleton`. The scope defines if and how long the returned
                    dependency will be cached. See :py:class:`~.core.container.Scope`.
                    Defaults to :py:meth:`~.core.container.Scope.singleton`.
            """
            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring must be a Wiring or None, "
                                f"not {type(wiring)}")

            super().__init__(wiring=wiring,
                             scope=validated_scope(scope,
                                                   singleton,
                                                   default=Scope.singleton()),
                             parameters=validated_parameters(parameters))

        def copy(self,
                 *,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 singleton: Union[bool, Copy] = Copy.IDENTICAL,
                 scope: Union[Optional[Scope], Copy] = Copy.IDENTICAL,
                 parameters: Union[Optional[Iterable[str]], Copy] = Copy.IDENTICAL,
                 ) -> 'Factory.Conf':
            """
            Copies current configuration and overrides only specified arguments.
            Accepts the same arguments as :py:meth:`.__init__`
            """
            if not (singleton is Copy.IDENTICAL or scope is Copy.IDENTICAL):
                raise TypeError("Use either singleton or scope argument, not both.")
            if isinstance(singleton, bool):
                scope = Scope.singleton() if singleton else None
            return Copy.immutable(self,
                                  wiring=wiring,
                                  scope=scope,
                                  parameters=parameters)

    __antidote__: Conf = Conf()
    """
    Configuration of the factory. Defaults to wire :py:meth:`.__init__` and
    :py:meth:`.__call__`.
    """


@overload
def factory(f: F,  # noqa: E704  # pragma: no cover
            *,
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            ) -> FactoryProtocol[F]: ...


@overload
def factory(*,  # noqa: E704  # pragma: no cover
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            ) -> Callable[[F], FactoryProtocol[F]]: ...


@API.public
def factory(f: F = None,
            *,
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            ) -> Union[FactoryProtocol[F], Callable[[F], FactoryProtocol[F]]]:
    """
    Registers a factory which provides as single dependency, defined through the return
    type annotation.

    .. doctest:: factory

        >>> from antidote import factory
        >>> class Database:
        ...     pass
        >>> @factory
        ... def load_db() -> Database:
        ...     return Database()

    To retrieve the dependency from Antidote you need to use a specific syntax
    :code:`dependency @ factory` as presented in the following examples. The goal of it is
    twofold:

    - Ensure that the factory is loaded whenever you require the dependency.
    - Better maintainability as you know *where* the dependency comes from.

    .. doctest:: factory

        >>> from antidote import world, inject
        >>> world.get(Database @ load_db)  # treated as `object` by Mypy
        <Database ...>
        >>> # With Mypy casting
        ... world.get[Database](Database @ load_db)
        <Database ...>
        >>> # Concise Mypy casting
        ... world.get[Database] @ load_db
        <Database ...>
        >>> @inject([Database @ load_db])
        ... def f(db: Database):
        ...     pass

    Or with annotated type hints:

    .. doctest:: factory

        >>> from typing import Annotated
        ... # from typing_extensions import Annotated # Python < 3.9
        >>> from antidote import From
        >>> @inject
        ... def f(db: Annotated[Database, From(load_db)]):
        ...     pass

    The factory returns a singleton by default and is automatically injected, so
    you can use annotated type hints with it:

    .. doctest:: factory

        >>> # Singleton by default
        ... world.get[Database] @ load_db is world.get[Database] @ load_db
        True
        >>> class Session:
        ...     pass
        >>> @factory(singleton=False)
        ... def session_gen(db: Annotated[Database, From(load_db)]) -> Session:
        ...     return Session()
        >>> world.get[Session] @ session_gen is world.get[Session] @ session_gen
        False

    .. note::

        If you need a stateful factory or want to implement a complex one prefer using
        :py:class:`.Factory` instead.

    Args:
        f: Callable which builds the dependency.
        singleton: Whether the returned dependency  is a singleton or not. If yes,
            the factory will be called at most once and the result re-used. Mutually
            exclusive with :code:`scope`. Defaults to :py:obj:`True`.
        scope: Scope of the returned dependency. Mutually exclusive with
            :code:`singleton`. The scope defines if and how long the returned dependency
            will be cached. See :py:class:`~.core.container.Scope`. Defaults to
            :py:meth:`~.core.container.Scope.singleton`.
        tags: Iterable of :py:class:`~.._providers.tag.Tag` applied to the provided
            dependency.

    Returns:
        The factory or the function decorator.

    """
    scope = validated_scope(scope, singleton, default=Scope.singleton())

    @inject
    def register_factory(func: F,
                         factory_provider: Provide[FactoryProvider] = None
                         ) -> FactoryProtocol[F]:
        assert factory_provider is not None

        if not (inspect.isfunction(func)
                or (is_wrapper(func)
                    and inspect.isfunction(func.__wrapped__))):  # type: ignore
            raise TypeError(f"{func} is not a function")

        output = get_type_hints(func).get('return')
        if output is None:
            raise ValueError("A return type hint is necessary. "
                             "It is used a the dependency.")
        if not inspect.isclass(output):
            raise TypeError(f"The return type hint is expected to be a class, "
                            f"not {type(output)}.")

        try:
            func = inject(func)
        except DoubleInjectionError:
            pass

        dependency = factory_provider.register(factory=func,
                                               scope=scope,
                                               output=output)

        return cast(FactoryProtocol[F], FactoryWrapper(func, dependency))

    return f and register_factory(f) or register_factory



"""
Testing utilities with Antidote. Also used by Antidote itself.
"""
from . import override
from ._methods import clone, empty, factory, maybe_provide_from, new, singleton

__all__ = ['clone', 'empty', 'factory', 'maybe_provide_from', 'new', 'singleton']



from .core.exceptions import (AntidoteError, DependencyCycleError,
                              DependencyInstantiationError, DependencyNotFoundError,
                              DoubleInjectionError, DuplicateDependencyError,
                              FrozenWorldError)

__all__ = [
    'AntidoteError',
    'DependencyCycleError',
    'DependencyInstantiationError',
    'DependencyNotFoundError',
    'DuplicateDependencyError',
    'FrozenWorldError',
    'DoubleInjectionError'
]



"""
Antidote has a global container which is managed in this module.
"""
import threading
from contextlib import contextmanager
from typing import Callable, Iterator, Optional

from ..core.container import OverridableRawContainer, RawContainer

__container: Optional[RawContainer] = None
__container_lock = threading.RLock()


def current_container() -> RawContainer:
    assert __container is not None
    return __container


def current_overridable_container() -> OverridableRawContainer:
    assert __container is not None
    if not isinstance(__container, OverridableRawContainer):
        raise RuntimeError("Current world does not support overrides. "
                           "Consider using world.test.clone(override=True)")
    return __container


# Used only for tests
def reset() -> None:
    global __container
    __container = None


def init() -> None:
    global __container
    if __container is None:
        with __container_lock:
            if __container is None:
                from antidote._internal.world import new_container
                __container = new_container()


@contextmanager
def override(create: Callable[[RawContainer], RawContainer]) -> Iterator[None]:
    global __container
    with __container_lock:
        assert __container is not None
        old = __container
        try:
            __container = create(old)
            yield
        finally:
            __container = old



from typing import Callable, FrozenSet, Iterable, Optional, TypeVar, Union, cast, overload

from ._compatibility.typing import final
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from ._service import ServiceMeta, ABCServiceMeta
from ._utils import validated_parameters
from .core import Scope, Wiring, WithWiringMixin
from .core.exceptions import DuplicateDependencyError
from .utils import validated_scope

C = TypeVar('C', bound=type)


@API.public
class Service(metaclass=ServiceMeta, abstract=True):
    """
    .. note::

        If you encounter conflicts with :py:class:`.Service` metaclass, consider using
        the class decorator :py:func:`.service` instead. You may also use
        :py:class:`.ABCService` if you only need to be compatible with the
        :py:class:`abc.ABC` class.

    Defines subclasses as services:

    .. doctest:: service_class

        >>> from antidote import Service, world, inject
        >>> class Database(Service):
        ...     pass

    The service can then be retrieved by its class:

    .. doctest:: service_class

        >>> world.get(Database)  # treated as `object` by Mypy
        <Database ...>
        >>> # With Mypy casting
        ... world.get[Database]()
        <Database ...>
        >>> @inject([Database])
        ... def f(db: Database):
        ...     pass

    Or with annotated type hints:

    .. doctest:: service_class

        >>> from antidote import Provide
        >>> @inject
        ... def f(db: Provide[Database]):
        ...     pass

    All methods are injected by default and the service itself is a singleton.
    All of this can be configured with :py:attr:`.__antidote__`:

    .. doctest:: service_class

        >>> # Singleton by default
        ... world.get[Database]() is world.get[Database]()
        True
        >>> class Session(Service):
        ...     __antidote__ = Service.Conf(singleton=False)
        ...
        ...     def __init__(self, db: Provide[Database]):
        ...         self.db = db
        ...
        >>> world.get[Session]() is world.get[Session]()
        False

    You may also parameterize the service. Parameters will be passed on to
    :code:`__init__()`:

    .. doctest:: service_class

        >>> class Database(Service):
        ...     __antidote__ = Service.Conf(parameters=['host'])
        ...
        ...     def __init__(self, host: str):
        ...         self.host = host
        ...
        ...     @classmethod
        ...     def hosted(cls, host: str) -> object:
        ...         return cls.parameterized(host=host)
        ...
        >>> test_db = world.get[Database](Database.hosted('test'))
        >>> test_db.host
        'test'
        >>> # The factory returns a singleton so our test_session will also be one
        ... world.get[Database](Database.hosted('test')) is test_db
        True

    """
    __slots__ = ()

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable service configuration. To change parameters on a existing instance, use
        either method :py:meth:`.copy` or
        :py:meth:`.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'scope', 'parameters')
        wiring: Optional[Wiring]
        scope: Optional[Scope]
        parameters: Optional[FrozenSet[str]]

        @property
        def singleton(self) -> bool:
            return self.scope is Scope.singleton()

        def __init__(self,
                     *,
                     wiring: Optional[Wiring] = Wiring(),
                     singleton: bool = None,
                     scope: Optional[Scope] = Scope.sentinel(),
                     parameters: Iterable[str] = None):
            """
            Args:
                wiring: Wiring to be applied on the service. By default only
                    :code:`__init__()` will be wired. Unless :py:attr:`.factory` is class
                    method name, in which case only the latter would be wired. To
                    deactivate any wiring at all use :py:obj:`None`.
                singleton: Whether the service is a singleton or not. A singleton is
                    instantiated only once. Mutually exclusive with :code:`scope`.
                    Defaults to :py:obj:`True`
                scope: Scope of the service. Mutually exclusive with :code:`singleton`.
                    The scope defines if and how long the service will be cached. See
                    :py:class:`~.core.container.Scope`. Defaults to
                    :py:meth:`~.core.container.Scope.singleton`.
            """
            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring must be a Wiring or None, not {type(wiring)}")

            super().__init__(wiring=wiring,
                             scope=validated_scope(scope,
                                                   singleton,
                                                   default=Scope.singleton()),
                             parameters=validated_parameters(parameters))

        def copy(self,
                 *,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 singleton: Union[bool, Copy] = Copy.IDENTICAL,
                 scope: Union[Optional[Scope], Copy] = Copy.IDENTICAL,
                 parameters: Union[Optional[Iterable[str]], Copy] = Copy.IDENTICAL,
                 ) -> 'Service.Conf':
            """
            Copies current configuration and overrides only specified arguments.
            Accepts the same arguments as :py:meth:`.__init__`
            """
            if not (singleton is Copy.IDENTICAL or scope is Copy.IDENTICAL):
                raise TypeError("Use either singleton or scope argument, not both.")
            if isinstance(singleton, bool):
                scope = Scope.singleton() if singleton else None
            return Copy.immutable(self, wiring=wiring, scope=scope, parameters=parameters)

    __antidote__: Conf = Conf()
    """
    Configuration of the service. Defaults to wire :py:meth:`.__init__`.
    """


@API.public
class ABCService(Service, metaclass=ABCServiceMeta, abstract=True):
    """
    This class only purpose is to facilitate the use of a abstract parent class, relying
    on :py:class:`abc.ABC`, with :py:class:`.Service`.

    .. doctest:: abc_service_class

        >>> from abc import ABC, abstractmethod
        >>> from antidote import ABCService, world
        >>> class AbstractClass(ABC):
        ...     @abstractmethod
        ...     def hello(self) -> str:
        ...         pass
        >>> class MyService(AbstractClass, ABCService):
        ...     def hello(self) -> str:
        ...         return "world"
        >>> world.get[MyService]().hello()
        'world'
    """
    __slots__ = ()


@overload
def service(klass: C,  # noqa: E704  # pragma: no cover
            *,
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            wiring: Optional[Wiring] = Wiring()
            ) -> C: ...


@overload
def service(*,  # noqa: E704  # pragma: no cover
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            wiring: Optional[Wiring] = Wiring()
            ) -> Callable[[C], C]: ...


@API.public
def service(klass: C = None,
            *,
            singleton: bool = None,
            scope: Optional[Scope] = Scope.sentinel(),
            wiring: Optional[Wiring] = Wiring()
            ) -> Union[C, Callable[[C], C]]:
    """
    Defines the decorated class as a service. To be used when you cannot inherit
    :py:class:`.Service`

    .. doctest:: service_decorator

        >>> from antidote import service
        >>> @service
        ... class CannotInheritService:
        ...     pass

    Like :py:class:`.Service` it'll automatically wire the class.

    .. doctest:: service_decorator

        >>> from antidote import Provide, world
        >>> @service
        ... class Test:
        ...     def __init__(self, cis: Provide[CannotInheritService]):
        ...         self.cis = cis
        >>> world.get[Test]().cis
        <CannotInheritService object at ...>

    .. note::

        If your wish to declare to register an external class to Antidote, prefer using
        a factory with either :py:class:`~.factory.Factory` or
        :py:func:`~.factory.factory`.

    Args:
        klass: Class to register as a dependency. It will be instantiated  only when
            requested.
        singleton: Whether the service is a singleton or not. A singleton is
            instantiated only once. Mutually exclusive with :code:`scope`.
            Defaults to :py:obj:`True`
        scope: Scope of the service. Mutually exclusive with :code:`singleton`.
            The scope defines if and how long the service will be cached. See
            :py:class:`~.core.container.Scope`. Defaults to
            :py:meth:`~.core.container.Scope.singleton`.
        wiring: :py:class:`.Wiring` to be used on the class. By defaults will apply
            a simple :py:func:`.inject` on all methods, so only annotated type hints are
            taken into account. Can be deactivated by specifying :py:obj:`None`.

    Returns:
        The decorated class, unmodified, if specified or the class decorator.

    """
    scope = validated_scope(scope, singleton, default=Scope.singleton())
    if wiring is not None and not isinstance(wiring, Wiring):
        raise TypeError(f"wiring must be a Wiring or None, not a {type(wiring)!r}")

    def reg(cls: C) -> C:
        from ._service import _configure_service

        if issubclass(cls, Service):
            raise DuplicateDependencyError(f"{cls} is already defined as a dependency "
                                           f"by inheriting {Service}")

        if wiring is not None:
            wiring.wire(cls)

        _configure_service(cls, conf=Service.Conf(wiring=None, scope=scope))

        return cast(C, cls)

    return klass and reg(klass) or reg



import inspect
from typing import (Any, Callable, Dict, Iterator, List, Sequence, Set, Union)

from .utils import FinalImmutable
from .._compatibility.typing import get_type_hints


class Argument(FinalImmutable):
    __slots__ = ('name', 'has_default', 'type_hint', 'type_hint_with_extras')
    name: str
    has_default: bool
    type_hint: Any
    type_hint_with_extras: Any

    def __repr__(self) -> str:
        return f"Argument({self}, extras={self.type_hint_with_extras})"

    def __str__(self) -> str:
        type_hint = getattr(self.type_hint, "__name__", repr(self.type_hint))
        common = f'{self.name}:{type_hint}'
        return common + " = ?" if self.has_default else common


class Arguments:
    """ Used when generating the injection wrapper """
    arguments: Sequence[Argument]
    has_var_positional: bool
    has_var_keyword: bool
    has_self: bool
    without_self: 'Arguments'
    __name_to_argument: Dict[str, Argument]

    @classmethod
    def from_callable(cls, f: Union[Callable[..., object], staticmethod, classmethod]
                      ) -> 'Arguments':
        if not (callable(f) or isinstance(f, (staticmethod, classmethod))):
            raise TypeError(f"func must be a callable or a static/class-method. "
                            f"Not a {type(f)}")
        return cls._build(
            f.__func__ if isinstance(f, (staticmethod, classmethod)) else f,
            is_unbound_method(f)  # doing it before un-wrapping.
        )

    @classmethod
    def _build(cls, func: Callable[..., object], unbound_method: bool) -> 'Arguments':
        arguments: List[Argument] = []
        has_var_positional = False
        has_var_keyword = False

        # typing is used, as lazy evaluation is not done properly with Signature.
        type_hints = get_type_hints(func)
        extra_type_hints = get_type_hints(func, include_extras=True)

        for name, parameter in inspect.signature(func).parameters.items():
            if parameter.kind is parameter.VAR_POSITIONAL:
                has_var_positional = True
            elif parameter.kind is parameter.VAR_KEYWORD:
                has_var_keyword = True
            else:
                arguments.append(Argument(
                    name=name,
                    has_default=parameter.default is not parameter.empty,
                    type_hint=type_hints.get(name),
                    type_hint_with_extras=extra_type_hints.get(name)
                ))

        return Arguments(arguments=tuple(arguments),
                         has_var_positional=has_var_positional,
                         has_var_keyword=has_var_keyword,
                         has_self=unbound_method)

    def __init__(self,
                 arguments: Sequence[Argument],
                 has_var_positional: bool,
                 has_var_keyword: bool,
                 has_self: bool):
        self.arguments = arguments
        self.__name_to_argument = dict(((arg.name, arg)
                                        for arg in arguments))
        self.has_var_positional = has_var_positional
        self.has_var_keyword = has_var_keyword
        self.has_self = has_self

        if has_self:
            self.without_self = Arguments(self.arguments[1:], self.has_var_positional,
                                          self.has_var_keyword, has_self=False)
        else:
            self.without_self = self

    @property
    def arg_names(self) -> Set[str]:
        return set(self.__name_to_argument.keys())

    def __repr__(self) -> str:
        args = [str(arg) for arg in self.__name_to_argument.values()]
        if self.has_var_positional:
            args.append("*args")
        if self.has_var_keyword:
            args.append("**kwargs")

        return f"Arguments({', '.join(args)})"

    def __getitem__(self, index: Union[str, int]) -> Argument:
        if isinstance(index, str):
            return self.__name_to_argument[index]
        elif isinstance(index, int):
            return self.arguments[index]
        else:
            raise TypeError(f"Unsupported index {index!r}")

    def __contains__(self, name: str) -> bool:
        return name in self.__name_to_argument

    def __len__(self) -> int:
        return len(self.arguments)

    def __iter__(self) -> Iterator[Argument]:
        return iter(self.arguments)


def is_unbound_method(func: Union[Callable[..., object], staticmethod, classmethod]
                      ) -> bool:
    """
    Methods and nested function will have a different __qualname__ than
    __name__ (See PEP-3155).

    >>> class A:
    ...     def f(self):
    ...         pass
    >>> A.f.__qualname__
    'A.f'

    This helps us differentiate method defined in a module and those for a class.
    """
    if isinstance(func, staticmethod):
        return False

    if isinstance(func, classmethod):
        func = func.__func__

    return (func.__qualname__ != func.__name__  # not top level
            # not a bound method (self/cls already bound)
            and not inspect.ismethod(func)
            # not nested function
            and not func.__qualname__[:-len(func.__name__)].endswith("<locals>."))



import pytest

from antidote import world
from antidote._providers import (FactoryProvider, IndirectProvider,
                                 LazyProvider, ServiceProvider)
from antidote.core.container import RawProvider


@pytest.fixture(params=[
    FactoryProvider,
    ServiceProvider,
    LazyProvider,
    IndirectProvider
])
def provider(request):
    return request.param()


def test_unknown_dependency(provider: RawProvider):
    assert world.test.maybe_provide_from(provider, object()) is None
    assert provider.maybe_debug(object()) is None
    assert not provider.exists(object())


def test_clone(provider: RawProvider):
    assert isinstance(provider.clone(False), type(provider))



# flake8: noqa
# Ignoring F811 for multiple definitions
from typing import Optional, TypeVar

from antidote._compatibility.typing import Annotated


def test_readme_simple():
    # from typing import Annotated
    # # from typing_extensions import Annotated # Python < 3.9
    from antidote import Service, inject, Provide

    class Database(Service):
        pass

    @inject
    def f(db: Provide[Database]):
        pass

    f()  # works !

    ######################

    f(Database())

    ######################

    @inject
    def f(db: Provide[Database]):
        pass

    ######################

    @inject([Database])
    def f(db):
        pass

    ######################

    @inject({'db': Database})
    def f(db):
        pass

    ######################

    # All class type hints are treated as dependencies
    @inject(auto_provide=True)
    def f(db: Database):
        pass

    ######################

    from antidote import inject, Service, Constants, const

    class Config(Constants):
        DB_HOST = const('localhost:5432')

    class Database(Service):
        @inject([Config.DB_HOST])  # self is ignored when specifying a list
        def __init__(self, host: str):
            self._host = host

    @inject({'db': Database})
    def f(db: Database):
        pass

    f()  # yeah !

    ######################

    from antidote import world

    # Retrieve dependencies by hand, in tests typically
    world.get(Config.DB_HOST)
    world.get[str](Config.DB_HOST)  # with type hint
    world.get[Database]()  # omit dependency if it's the type hint itself

    ######################

    @inject
    def f(db: Provide[Database] = None):
        # Used to tell Mypy that `db` is optional but must be either injected or given.
        assert db is not None
        pass


def test_readme():

    # from a library
    class ImdbAPI:
        def __init__(self, host: str, port: int, api_key: str):
            pass

    ######################

    # movie.py
    class MovieDB:
        """ Interface """

        def get_best_movies(self):
            pass

    ######################

    # config.py
    from antidote import Constants, const

    class Config(Constants):
        # with str/int/float, the type hint is enforced. Can be removed or extend to
        # support Enums.
        IMDB_HOST = const[str]('imdb.host')
        IMDB_PORT = const[int]('imdb.port')
        IMDB_API_KEY = const('imdb.api_key')

        def __init__(self):
            self._raw_conf = {
                'imdb': {
                    'host': 'dummy_host',
                    'api_key': 'dummy_api_key',
                    'port': '80'
                }
            }

        def provide_const(self, name: str, arg: str):
            root, key = arg.split('.')
            return self._raw_conf[root][key]

    ######################

    # current_movie.py
    # Code implementing/managing MovieDB
    from antidote import factory, inject, Service, implementation
    # from config import Config

    # Provides ImdbAPI, as defined by the return type annotation.
    @factory
    @inject([Config.IMDB_HOST, Config.IMDB_PORT, Config.IMDB_API_KEY])
    def imdb_factory(host: str, port: int, api_key: str) -> ImdbAPI:
        # Here host = Config().provide_const('IMDB_HOST', 'imdb.host')
        return ImdbAPI(host=host, port=port, api_key=api_key)

    class IMDBMovieDB(MovieDB, Service):
        __antidote__ = Service.Conf(singleton=False)  # New instance each time

        @inject({'imdb_api': ImdbAPI @ imdb_factory})
        def __init__(self, imdb_api: ImdbAPI):
            self._imdb_api = imdb_api

        def get_best_movies(self):
            pass

    @implementation(MovieDB)
    def current_movie_db() -> object:
        return IMDBMovieDB  # dependency to be provided for MovieDB

    ######################

    # current_movie.py
    # Code implementing/managing MovieDB
    from antidote import factory, Service, Get, From
    # from typing import Annotated
    # # from typing_extensions import Annotated # Python < 3.9
    # from config import Config

    @factory
    def imdb_factory(host: Annotated[str, Get(Config.IMDB_HOST)],
                     port: Annotated[int, Get(Config.IMDB_PORT)],
                     api_key: Annotated[str, Get(Config.IMDB_API_KEY)]
                     ) -> ImdbAPI:
        return ImdbAPI(host, port, api_key)

    class IMDBMovieDB(MovieDB, Service):
        __antidote__ = Service.Conf(singleton=False)

        def __init__(self, imdb_api: Annotated[ImdbAPI, From(imdb_factory)]):
            self._imdb_api = imdb_api

        def get_best_movies(self):
            pass

    ######################

    # main.py
    # from movie import MovieDB
    # from current_movie import current_movie_db

    @inject([MovieDB @ current_movie_db])
    def main(movie_db: MovieDB = None):
        assert movie_db is not None  # for Mypy, to understand that movie_db is optional
        pass

    # Or with annotated type hints
    @inject
    def main(movie_db: Annotated[MovieDB, From(current_movie_db)]):
        pass

    main()

    ######################

    conf = Config()
    main(IMDBMovieDB(imdb_factory(
        # constants can be retrieved directly on an instance
        host=conf.IMDB_HOST,
        port=conf.IMDB_PORT,
        api_key=conf.IMDB_API_KEY,
    )))

    ######################

    from antidote import world

    # Clone current world to isolate it from the rest
    with world.test.clone():
        # Override the configuration
        world.test.override.singleton(Config.IMDB_HOST, 'other host')
        main()

    ######################

    world.debug(main)



import pytest

from antidote.core.container import (DependencyValue, OverridableRawContainer,
                                     RawContainer, Scope)
from antidote.exceptions import (DependencyCycleError, DependencyInstantiationError,
                                 DependencyNotFoundError)
from .utils import DummyProvider


@pytest.fixture
def original_container():
    c = RawContainer()
    c.add_provider(DummyProvider)
    c.get(DummyProvider).data = {'name': 'Antidote'}
    return c


@pytest.fixture
def container(original_container):
    return original_container.clone(keep_singletons=True,
                                    keep_scopes=False)


def test_override_singletons(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"

    container.override_singletons({'name': 'different'})
    assert container.get('name') == 'different'
    container.override_singletons({'name': 'different-v2'})
    assert container.get('name') == 'different-v2'
    assert container.provide('name') == DependencyValue('different-v2',
                                                        scope=Scope.singleton())


def test_override_factory(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"

    container.override_factory('name', factory=lambda: 'different', scope=None)
    assert container.get('name') == 'different'
    container.override_factory('name', factory=lambda: 'different-v2', scope=None)
    assert container.get('name') == 'different-v2'
    assert container.provide('name') == DependencyValue('different-v2',
                                                        scope=None)

    # overriding previous factory which had no scope
    container.override_factory('name', factory=lambda: object(), scope=Scope.singleton())
    value = container.get('name')
    assert not isinstance(value, str)
    assert container.get('name') is value

    # override still works
    container.override_factory('name', factory=lambda: object(), scope=Scope.singleton())
    assert container.get('name') is not value

    container.override_singletons({'name': 'Hello'})
    assert container.get('name') == 'Hello'


def test_override_provider(container: OverridableRawContainer):
    assert container.get('name') == "Antidote"

    container.override_provider(
        lambda x: DependencyValue('different') if x == 'name' else None)
    assert container.get('name') == 'different'
    container.override_provider(
        lambda x: DependencyValue('different-v2') if x == 'name' else None)
    assert container.get('name') == 'different-v2'
    assert container.provide('name') == DependencyValue('different-v2',
                                                        scope=None)

    with pytest.raises(DependencyNotFoundError):
        container.get('x')  # new provider does not bring any other value

    # overriding previous provider which had no scope
    container.override_provider(
        lambda x: DependencyValue(object(),
                                  scope=Scope.singleton()) if x == 'name' else None)
    value = container.get('name')
    assert not isinstance(value, str)
    assert container.get('name') is value

    container.override_provider(
        lambda x: DependencyValue(object(),
                                  scope=Scope.singleton()) if x == 'name' else None)
    # didn't change anything, because it's now a singleton.
    assert container.get('name') is value

    container.override_singletons({'name': 'Hello'})
    assert container.get('name') == 'Hello'


def test_errors(container: OverridableRawContainer):
    container.override_factory('cycle', factory=lambda: container.get('cycle'),
                               scope=None)
    with pytest.raises(DependencyCycleError):
        container.get('cycle')

    def factory():
        raise RuntimeError()

    container.override_factory('error', factory=factory, scope=None)
    with pytest.raises(DependencyInstantiationError):
        container.get('error')



from typing import (Any, Dict, Generic, Hashable, Optional, TYPE_CHECKING, Tuple, Type,
                    TypeVar, Union, cast, overload)

from ._compatibility.typing import final, Protocol
from ._internal import API
from ._internal.utils import AbstractMeta, Default, FinalImmutable, FinalMeta, debug_repr
from ._internal.utils.immutable import Immutable, ImmutableGenericMeta
from ._providers.lazy import Lazy
from .core import Container, DependencyDebug, DependencyValue, Scope

T = TypeVar('T')

if TYPE_CHECKING:
    from .constants import Constants


# TODO: Once Python 3.6 support drops, fix this.
# We're lying to Mypy here. That's not how the actual descriptor, even though it's
# somewhat close. But inheriting Generic implies not being final anymore in Python 3.6,
# until PEP 560, and internally there's no need for Generic.
class Const(Generic[T]):
    __slots__ = ()

    @overload
    def __get__(self,  # noqa: E704
                instance: 'Constants',
                owner: 'Type[Constants]') -> T: ...  # pragma: no cover

    @overload
    def __get__(self,  # noqa: E704
                instance: None,
                owner: 'Type[Constants]') -> 'Const[T]': ...  # pragma: no cover

    def __get__(self,
                instance: 'Optional[Constants]',
                owner: 'Type[Constants]') -> object:  # pragma: no cover
        pass


@API.private
@final
class MakeConst(metaclass=FinalMeta):
    def __call__(self,
                 __arg: Optional[object] = None,
                 *,
                 default: Any = Default.sentinel) -> Const[object]:
        # Not true yet, but will be changed by ConstantsMeta
        return cast(Const[object], LazyConstToDo(__arg, None, default))

    def __getitem__(self, tpe: Type[T]) -> 'MakeTypedConst[T]':
        return MakeTypedConst(tpe)


@API.private
@final
class MakeTypedConst(Immutable, Generic[T], metaclass=ImmutableGenericMeta):
    __slots__ = ('__type',)
    __type: Type[T]

    def __call__(self,
                 __arg: Optional[object] = None,
                 *,
                 default: Union[T, Default] = Default.sentinel) -> Const[T]:
        if not isinstance(default, (self.__type, Default)):
            raise TypeError(f"default is not an instance of {self.__type}, "
                            f"but {type(default)}")

        # Not true yet, but will be changed by ConstantsMeta
        return cast(Const[T], LazyConstToDo(__arg, self.__type, default))


@API.private
@final
class LazyConstToDo(FinalImmutable):
    __slots__ = ('arg', 'type_', 'default')
    arg: Optional[object]
    type_: Optional[type]
    default: object


@API.private
class ConstantsMeta(AbstractMeta):
    def __new__(mcs: 'Type[ConstantsMeta]',
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                **kwargs: object
                ) -> 'ConstantsMeta':
        cls = cast(
            ConstantsMeta,
            super().__new__(mcs, name, bases, namespace, **kwargs)  # type: ignore
        )
        if not kwargs.get('abstract'):
            _configure_constants(cls)
        return cls


@API.private
def _configure_constants(cls: ConstantsMeta) -> None:
    from .constants import Constants
    from .service import service

    conf = getattr(cls, '__antidote__', None)
    if not isinstance(conf, Constants.Conf):
        raise TypeError(f"Constants configuration (__antidote__) is expected to be a "
                        f"{Constants.Conf}, not a {type(conf)}")

    cls = service(cls, singleton=True, wiring=conf.wiring)
    for name, v in list(cls.__dict__.items()):
        if isinstance(v, LazyConstToDo):
            setattr(cls,
                    name,
                    LazyConstDescriptor(
                        name=name,
                        dependency=cls,
                        method_name=Constants.provide_const.__name__,
                        arg=v.arg,
                        default=v.default,
                        type_=v.type_ or object,
                        auto_cast=v.type_ is not None and v.type_ in conf.auto_cast))


@API.private
@final
class LazyConstDescriptor(FinalImmutable):
    __slots__ = ('name', 'dependency', 'method_name', 'arg', 'default', 'type_',
                 'auto_cast', '_cache')
    name: str
    dependency: Hashable
    method_name: str
    arg: object
    default: object
    type_: type
    auto_cast: bool
    _cache: str

    def __init__(self,
                 *,
                 name: str,
                 dependency: Hashable,
                 method_name: str,
                 arg: object,
                 default: object,
                 type_: type,
                 auto_cast: bool
                 ):
        assert isinstance(default, (Default, type_))
        super().__init__(
            name=name,
            dependency=dependency,
            method_name=method_name,
            arg=arg,
            default=default,
            type_=type_,
            auto_cast=auto_cast,
            _cache=f"__antidote_dependency_{hex(id(self))}"
        )

    def __get__(self, instance: object, owner: type) -> object:
        if instance is None:
            try:
                return getattr(owner, self._cache)
            except AttributeError:
                dependency = LazyConst(self)
                setattr(owner, self._cache, dependency)
                return dependency
        try:
            value = getattr(instance, self.method_name)(name=self.name,
                                                        arg=self.arg)
        except LookupError:
            if self.default is not Default.sentinel:
                return self.default
            raise

        if self.auto_cast:
            value = self.type_(value)

        if not isinstance(value, self.type_):
            raise TypeError(f"Constant {self.name} is not an instance of {self.type_}, "
                            f"but {type(value)}")

        return value


@API.private
@final
class LazyConst(FinalImmutable, Lazy):
    __slots__ = ('descriptor',)
    descriptor: LazyConstDescriptor

    def __init__(self, descriptor: LazyConstDescriptor) -> None:
        super().__init__(descriptor=descriptor)

    def debug_info(self) -> DependencyDebug:
        descriptor = cast(LazyConstDescriptor, self.descriptor)
        cls = cast(type, descriptor.dependency)
        return DependencyDebug(f"{debug_repr(cls)}.{descriptor.name}",
                               scope=Scope.singleton(),
                               # TODO: Would be great if the first argument of the method
                               #       didn't show as unknown as it's always provided.
                               wired=[getattr(cls, descriptor.method_name), cls])

    def provide(self, container: Container) -> DependencyValue:
        # TODO: Waiting for a fix: https://github.com/python/mypy/issues/6910
        descriptor = cast(LazyConstDescriptor, self.descriptor)
        return DependencyValue(
            descriptor.__get__(
                container.get(descriptor.dependency),
                None  # type: ignore
            ),
            scope=Scope.singleton()
        )



"""
Cython version of the wrapper, doing the same thing but faster.
"""
# @formatter:off
import inspect
cimport cython
from cpython.object cimport PyObject_Call, PyObject_CallMethodObjArgs
from cpython.ref cimport PyObject

from antidote._internal.state cimport fast_get_container
from antidote.core.container cimport (DependencyResult, RawContainer)
from ..core.exceptions import DependencyNotFoundError

# @formatter:on

cdef extern from "Python.h":
    PyObject*PyTuple_GET_ITEM(PyObject *p, Py_ssize_t pos)
    Py_ssize_t PyTuple_GET_SIZE(PyObject *p)
    int PyDict_Contains(PyObject *p, PyObject *key) except -1
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    Py_ssize_t PyDict_Size(PyObject *p)
    PyObject*PyDict_Copy(PyObject *p)
    PyObject*PyDict_New() except NULL
    void Py_DECREF(PyObject *o)
    void Py_INCREF(PyObject *o)


compiled = True

cdef class Injection:
    cdef:
        str arg_name
        bint required
        object dependency

    def __repr__(self):
        return f"Injection(arg_name={self.arg_name!r}, " \
               f"required={self.required!r}, dependency={self.dependency!r})"

    def __init__(self, str arg_name, bint required, object dependency):
        self.arg_name = arg_name
        self.required = required
        self.dependency = dependency

cdef class InjectionBlueprint:
    cdef:
        tuple injections

    def __init__(self, tuple injections):
        self.injections = injections

    def is_empty(self):
        cdef:
            Injection injection
        for injection in self.injections:
            if injection.dependency is not None:
                return False
        return True

    def __repr__(self):
        return f"InjectionBlueprint({','.join(map(repr, self.injections))})"

def build_wrapper(InjectionBlueprint blueprint,
                  object wrapped,
                  bint skip_first = False):
    """
    Not relying on __init__ neither __cinit__ makes initialization faster which
    is especially important during __get__ as a InjectedBoundWrapper needs to
    be created *every* time.
    """
    cdef:
        InjectedWrapper wrapper
    if inspect.iscoroutinefunction(wrapped):
        wrapper = AsyncInjectedWrapper.__new__(AsyncInjectedWrapper)
    else:
        wrapper = SyncInjectedWrapper.__new__(SyncInjectedWrapper)
    wrapper.__wrapped__ = wrapped
    wrapper.__blueprint = blueprint
    wrapper.__injection_offset = 1 if skip_first else 0
    return wrapper

def get_wrapper_injections(wrapper):
    if not isinstance(wrapper, InjectedWrapper):
        raise TypeError(f"Argument must be an {InjectedWrapper}")

    return (<InjectedWrapper> wrapper).get_injections()

def is_wrapper(x):
    return isinstance(x, InjectedWrapper)

def get_wrapped(x):
    assert isinstance(x, InjectedWrapper)
    return x.__wrapped__

@cython.freelist(128)
cdef class InjectedWrapper:
    cdef:
        dict __dict__
        readonly object __wrapped__
        InjectionBlueprint __blueprint
        int __injection_offset
        object __weakref__

    cdef dict get_injections(self):
        cdef:
            Injection inj
        return {inj.arg_name: inj.dependency
                for inj in self.__blueprint.injections
                if inj.dependency is not None}

    def __getattr__(self, name):
        return getattr(self.__wrapped__, name)

# SYNC

cdef class SyncInjectedWrapper(InjectedWrapper):
    def __call__(self, *args, **kwargs):
        kwargs = <object> build_kwargs(<PyObject*> args,
                                       <PyObject*> kwargs,
                                       <PyObject*> self)
        Py_DECREF(<PyObject*> kwargs)
        return PyObject_Call(self.__wrapped__, args, kwargs)

    def __get__(self, instance, owner):
        cdef:
            SyncInjectedBoundWrapper bound_wrapper = \
                SyncInjectedBoundWrapper.__new__(SyncInjectedBoundWrapper)
        setup_bound_wrapper(<PyObject*> self,
                            <PyObject*> bound_wrapper,
                            <PyObject*> instance,
                            <PyObject*> owner)
        return bound_wrapper

cdef class SyncInjectedBoundWrapper(SyncInjectedWrapper):
    def __get__(self, instance, owner):
        return self

# ASYNC

cdef class AsyncInjectedWrapper(InjectedWrapper):
    async def __call__(self, *args, **kwargs):
        kwargs = <object> build_kwargs(<PyObject*> args,
                                       <PyObject*> kwargs,
                                       <PyObject*> self)
        Py_DECREF(<PyObject*> kwargs)
        return await PyObject_Call(self.__wrapped__, args, kwargs)

    def __get__(self, instance, owner):
        cdef:
            AsyncInjectedBoundWrapper bound_wrapper = \
                AsyncInjectedBoundWrapper.__new__(AsyncInjectedBoundWrapper)
        setup_bound_wrapper(<PyObject*> self,
                            <PyObject*> bound_wrapper,
                            <PyObject*> instance,
                            <PyObject*> owner)
        return bound_wrapper

cdef class AsyncInjectedBoundWrapper(AsyncInjectedWrapper):
    def __get__(self, instance, owner):
        return self

# Mixin

cdef inline setup_bound_wrapper(PyObject*wrapper,
                                PyObject*bound_wrapper,
                                PyObject*instance,
                                PyObject*owner):
    (<InjectedWrapper> bound_wrapper).__wrapped__ = \
        PyObject_CallMethodObjArgs((<InjectedWrapper> wrapper).__wrapped__,
                                    "__get__",
                                    instance,
                                    owner,
                                    NULL)
    (<InjectedWrapper> bound_wrapper).__blueprint = (<InjectedWrapper> wrapper).__blueprint
    if instance != <PyObject*>None:
        (<InjectedWrapper> bound_wrapper).__injection_offset = 1
    else:
        (<InjectedWrapper> bound_wrapper).__injection_offset = 0

cdef PyObject* build_kwargs(PyObject*args,
                            PyObject*original_kwargs,
                            PyObject*wrapper) except NULL:
    cdef:
        RawContainer container = fast_get_container()
        DependencyResult result
        PyObject*injection
        PyObject*arg_name
        PyObject*kwargs = original_kwargs
        PyObject*injections = <PyObject*> (<InjectedWrapper> wrapper).__blueprint.injections
        Py_ssize_t i
        Py_ssize_t offset = ((<InjectedWrapper> wrapper).__injection_offset
                             + PyTuple_GET_SIZE(<PyObject*> args))
        Py_ssize_t n = PyTuple_GET_SIZE(injections)

    if PyDict_Size(kwargs):
        for i in range(offset, n):
            injection = PyTuple_GET_ITEM(injections, i)
            if (<Injection> injection).dependency is not None:
                arg_name = <PyObject*> (<Injection> injection).arg_name
                if not PyDict_Contains(<PyObject*> kwargs, arg_name):
                    container.fast_get(<PyObject*> (<Injection> injection).dependency,
                                       &result)
                    if result.value:
                        if kwargs == original_kwargs:
                            kwargs = PyDict_Copy(original_kwargs)
                        PyDict_SetItem(<PyObject*> kwargs,
                                       arg_name,
                                       result.value)
                        Py_DECREF(result.value)

                    elif (<Injection> injection).required:
                        raise DependencyNotFoundError((<Injection> injection).dependency)
    else:
        for i in range(offset, n):
            injection = PyTuple_GET_ITEM(injections, i)
            if (<Injection> injection).dependency is not None:
                container.fast_get(<PyObject*> (<Injection> injection).dependency,
                                   &result)
                if result.value:
                    if kwargs == original_kwargs:
                        kwargs = PyDict_New()
                    PyDict_SetItem(
                        <PyObject*> kwargs,
                        <PyObject*> (<Injection> injection).arg_name,
                        result.value
                    )
                    Py_DECREF(result.value)
                elif (<Injection> injection).required:
                    raise DependencyNotFoundError((<Injection> injection).dependency)

    return kwargs



import collections.abc as c_abc
import enum
from typing import (Callable, FrozenSet, Hashable, Iterable, Optional, TypeVar,
                    Union,
                    overload)

from .injection import DEPENDENCIES_TYPE, validate_injection, AUTO_PROVIDE_TYPE
from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import Copy, FinalImmutable

C = TypeVar('C', bound=type)
_empty_set: FrozenSet[str] = frozenset()


class Methods(enum.Enum):
    ALL = enum.auto()


@API.public
@final
class Wiring(FinalImmutable):
    """
    Defines how a class should be wired, meaning if/how/which methods are injected. This
    class is intended to be used by configuration objects. If you just want to wire a
    single class, consider using the class decorator :py:func:`~.wire` instead. There are
    two purposes:

    - providing a default injection which can be overridden either by changing the wiring
      or using `@inject` when using :code:`attempt_methods`.
    - wiring of multiple methods with similar dependencies.

    Instances are immutable. If you want to change some parameters, typically defaults
    defined by Antidote, you'll need to rely on :py:meth:`~.copy`. It accepts the same
    arguments as :py:meth:`~.__init__` and overrides existing values.

    .. doctest:: core_Wiring

        >>> from antidote import Wiring, Service, Provide
        >>> wiring = Wiring(methods=['my_method'])
        >>> class Database(Service):
        ...     pass
        >>> @wiring.wire
        ... class Dummy:
        ...     def my_method(self, db: Provide[Database]) -> Database:
        ...         return db
        >>> Dummy().my_method()
        <Database ...>


    """
    __slots__ = ('methods', 'auto_provide', 'dependencies',
                 'raise_on_double_injection')
    methods: Union[Methods, FrozenSet[str]]
    """Method names that must be injected."""
    dependencies: DEPENDENCIES_TYPE
    auto_provide: Union[bool, FrozenSet[type], Callable[[type], bool]]
    raise_on_double_injection: bool

    def __init__(self,
                 *,
                 methods: Union[Methods, Iterable[str]] = Methods.ALL,
                 dependencies: DEPENDENCIES_TYPE = None,
                 auto_provide: AUTO_PROVIDE_TYPE = False,
                 raise_on_double_injection: bool = False) -> None:
        """
        Args:
            methods: Names of methods to be injected. If any of them is already injected,
                an error will be raised. Consider using :code:`attempt_methods` otherwise.
            dependencies: Propagated for every method to :py:func:`~.injection.inject`
            auto_provide: Propagated for every method to :py:func:`~.injection.inject`
        """
        if not isinstance(raise_on_double_injection, bool):
            raise TypeError(f"raise_on_double_injection must be a boolean, "
                            f"not {type(raise_on_double_injection)}")

        if isinstance(auto_provide, str) \
                or not (isinstance(auto_provide, (c_abc.Iterable, bool))
                        or callable(auto_provide)):
            raise TypeError(f"auto_provide must be an iterable of method names, "
                            f"not {type(auto_provide)}.")
        if isinstance(auto_provide, c_abc.Iterable):
            auto_provide = frozenset(auto_provide)

        if isinstance(methods, str) or not isinstance(methods, (c_abc.Iterable, Methods)):
            raise TypeError(f"methods must be an iterable of method names, "
                            f"not {type(methods)}.")
        elif isinstance(methods, c_abc.Iterable):
            methods = frozenset(methods)
            if not all(isinstance(method, str) for method in methods):
                raise TypeError("methods is expected to contain methods names (str)")

        if not isinstance(auto_provide, str) and isinstance(auto_provide,
                                                            c_abc.Iterable):
            auto_provide = frozenset(auto_provide)
        validate_injection(dependencies, auto_provide)

        super().__init__(methods=methods,
                         dependencies=dependencies,
                         auto_provide=auto_provide,
                         raise_on_double_injection=raise_on_double_injection)

    def copy(self,
             *,
             methods: Union[Methods, Iterable[str], Copy] = Copy.IDENTICAL,
             dependencies: Union[DEPENDENCIES_TYPE, Copy] = Copy.IDENTICAL,
             auto_provide: Union[AUTO_PROVIDE_TYPE, Copy] = Copy.IDENTICAL,
             raise_on_double_injection: Union[bool, Copy] = Copy.IDENTICAL
             ) -> 'Wiring':
        """
        Copies current wiring and overrides only specified arguments.
        Accepts the same arguments as :py:meth:`.__init__`
        """
        return Copy.immutable(self,
                              methods=methods,
                              dependencies=dependencies,
                              auto_provide=auto_provide,
                              raise_on_double_injection=raise_on_double_injection)

    def wire(self, __klass: C) -> C:
        """
        Used to wire a class with specified configuration.

        Args:
            __klass: Class to wired

        Returns:
            The same class object with specified methods injected.
        """
        from ._wiring import wire_class
        return wire_class(__klass, self)


@overload
def wire(__klass: C,  # noqa: E704  # pragma: no cover
         *,
         methods: Union[Methods, Iterable[str]] = Methods.ALL,
         dependencies: DEPENDENCIES_TYPE = None,
         auto_provide: AUTO_PROVIDE_TYPE = False,
         raise_on_double_injection: bool = False
         ) -> C: ...


@overload
def wire(*,  # noqa: E704  # pragma: no cover
         methods: Union[Methods, Iterable[str]] = Methods.ALL,
         dependencies: DEPENDENCIES_TYPE = None,
         auto_provide: AUTO_PROVIDE_TYPE = False,
         raise_on_double_injection: bool = False
         ) -> Callable[[C], C]: ...


@API.public
def wire(__klass: C = None,
         *,
         methods: Union[Methods, Iterable[str]] = Methods.ALL,
         dependencies: DEPENDENCIES_TYPE = None,
         auto_provide: AUTO_PROVIDE_TYPE = False,
         raise_on_double_injection: bool = False
         ) -> Union[C, Callable[[C], C]]:
    """
    Wire a class by injecting specified methods. This avoids repetition if similar
    dependencies need to be injected in different methods. Methods will only be wrapped
    if and only if Antidote may inject a dependency in it, like :py:func:`.inject`.

    Args:
        raise_on_double_injection: Whether an error should be raised if method is already
            injected. Defaults to :py:obj:`False`.
        __klass: Class to wire.
        methods: Names of methods that must be injected. Defaults to all method
        dependencies: Propagated for every method to :py:func:`~.injection.inject`.
        auto_provide: Propagated for every method to :py:func:`~.injection.inject`.

    Returns:
        Wired class or a class decorator.

    """
    wiring = Wiring(
        methods=methods,
        dependencies=dependencies,
        auto_provide=auto_provide,
        raise_on_double_injection=raise_on_double_injection
    )

    def wire_methods(cls: C) -> C:
        from ._wiring import wire_class
        return wire_class(cls, wiring)

    return __klass and wire_methods(__klass) or wire_methods


W = TypeVar('W', bound='WithWiringMixin')


@API.private  # Inheriting this class yourself is not part of the public API
class WithWiringMixin:
    __slots__ = ()

    wiring: Optional[Wiring]

    def copy(self: W, *, wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL) -> W:
        raise NotImplementedError()  # pragma: no cover

    @API.public
    def with_wiring(self: W,
                    *,
                    methods: Union[Methods, Iterable[str], Copy] = Copy.IDENTICAL,
                    dependencies: Union[DEPENDENCIES_TYPE, Copy] = Copy.IDENTICAL,
                    auto_provide: Union[AUTO_PROVIDE_TYPE, Copy] = Copy.IDENTICAL,
                    raise_on_double_injection: Union[bool, Copy] = Copy.IDENTICAL,
                    ) -> W:
        """
        Accepts the same arguments as :py:class:`.Wiring`. Its only purpose is to provide
        a easier way to change the wiring:

        .. code-block:: python

            conf.with_wiring(dependencies={})
            # instead of
            conf.copy(wiring=conf.wiring.copy(dependencies={}))

        """
        if self.wiring is None:
            return self.copy(wiring=Wiring().copy(
                methods=methods,
                dependencies=dependencies,
                auto_provide=auto_provide,
                raise_on_double_injection=raise_on_double_injection))
        else:
            return self.copy(wiring=self.wiring.copy(
                methods=methods,
                dependencies=dependencies,
                auto_provide=auto_provide,
                raise_on_double_injection=raise_on_double_injection))



import pytest

from antidote import world
from antidote._internal.world import new_container, LazyDependency
from antidote.core import Container


def test_new_container():
    assert isinstance(new_container(), Container)


def test_dependency():
    with world.test.empty():
        world.test.singleton('x', object())
        d = LazyDependency('x', object)
        assert d.unwrapped == 'x'
        assert d.get() is world.get('x')

    class A:
        pass

    with world.test.empty():
        world.test.singleton('a', A())
        world.test.singleton('x', object())

        assert LazyDependency('a', A).get() is world.get('a')

        with pytest.raises(TypeError):
            LazyDependency('x', A).get()



from typing import Callable, Hashable, Optional

import pytest

from antidote import world
from antidote._providers import ServiceProvider
from antidote.core import (Container, DependencyValue, StatelessProvider)
from antidote.core.exceptions import FrozenWorldError
from antidote.exceptions import DependencyNotFoundError
from .utils import DummyIntProvider


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


@pytest.mark.parametrize('context,keeps_singletons,strategy', [
    pytest.param(lambda: world.test.empty(),
                 False,
                 'empty',
                 id='empty'),
    pytest.param(lambda: world.test.new(),
                 False,
                 'new',
                 id='new'),
    pytest.param(lambda: world.test.clone(),
                 False,
                 'clone',
                 id='clone'),
    pytest.param(lambda: world.test.clone(keep_singletons=True),
                 True,
                 'clone',
                 id='clone-with-singletons')
])
def test_world(context: Callable, keeps_singletons: bool, strategy: str):
    class DummyFloatProvider(StatelessProvider[float]):
        def exists(self, dependency: Hashable) -> bool:
            return isinstance(dependency, float)

        def provide(self, dependency: float, container: Container
                    ) -> Optional[DependencyValue]:
            return DependencyValue(dependency ** 2)

    x = object()
    y = object()
    world.test.singleton("x", x)
    world.provider(DummyIntProvider)
    provider = world.get(DummyIntProvider)

    with context():
        if keeps_singletons:
            assert world.get('x') == x
        else:
            with pytest.raises(DependencyNotFoundError):
                world.get('x')

        if strategy == 'new':
            assert isinstance(world.get[DummyIntProvider](), DummyIntProvider)
            assert world.get(DummyIntProvider).original is None
        elif strategy == 'clone':
            assert isinstance(world.get(DummyIntProvider), DummyIntProvider)
            assert world.get(DummyIntProvider) is not provider
            assert world.get(DummyIntProvider).original is provider
        elif strategy == 'empty':
            with pytest.raises(DependencyNotFoundError):
                world.get(DummyIntProvider)
            with pytest.raises(DependencyNotFoundError):
                world.get(ServiceProvider)

        if strategy != 'clone':
            world.test.singleton("y", y)
            assert world.get('y') is y

            world.provider(DummyFloatProvider)
            assert world.get(1.2) == 1.2 ** 2

            s = world.scopes.new(name='dummy')
            world.scopes.reset(s)

    with pytest.raises(DependencyNotFoundError):
        world.get('y')

    with pytest.raises(DependencyNotFoundError):
        world.get(1.2)

    with pytest.raises(DependencyNotFoundError):
        world.get(DummyFloatProvider)


def test_clone_keep_singletons():
    world.test.singleton("singleton", 2)
    world.provider(DummyIntProvider)

    with world.test.clone(keep_singletons=True):
        assert world.get("singleton") == 2
        assert world.get(10) == 20

    with world.test.clone(keep_singletons=False):
        with pytest.raises(DependencyNotFoundError):
            world.get("singleton")
        assert world.get(10) == 20


@pytest.mark.parametrize('keep_singletons', [True, False])
@pytest.mark.parametrize('keep_scopes', [True, False])
def test_clone_restrictions(keep_singletons, keep_scopes):
    with world.test.clone(keep_singletons=keep_singletons,
                          keep_scopes=keep_scopes):
        with pytest.raises(FrozenWorldError):
            world.scopes.new(name="new scope")

        with pytest.raises(FrozenWorldError):
            world.provider(DummyIntProvider)

        with pytest.raises(FrozenWorldError):
            world.test.singleton('test', 1)


def test_deep_clone():
    world.test.singleton("test", 1)

    with world.test.clone(keep_singletons=True):
        with world.test.clone(keep_singletons=False):
            with pytest.raises(DependencyNotFoundError):
                world.get("test")


def test_empty():
    world.test.singleton("singleton", 2)

    with world.test.empty():
        world.test.singleton("a", 3)
        assert world.get("a") == 3
        with pytest.raises(DependencyNotFoundError):
            world.get("singleton")
        world.provider(DummyIntProvider)
        assert world.get(10) == 20
        with pytest.raises(DependencyNotFoundError):
            world.get(ServiceProvider)

    with pytest.raises(DependencyNotFoundError):
        world.get("a")

    with pytest.raises(DependencyNotFoundError):
        assert world.get(10)


def test_new():
    world.test.singleton("singleton", 2)
    world.provider(DummyIntProvider)
    assert world.get(10) == 20

    with world.test.new():
        world.test.singleton("a", 3)
        assert world.get("a") == 3
        with pytest.raises(DependencyNotFoundError):
            world.get("singleton")
        assert world.get(10) == 20

    with pytest.raises(DependencyNotFoundError):
        world.get("a")


def test_provide_from():
    provider = DummyIntProvider()
    assert world.test.maybe_provide_from(provider, 10) == DependencyValue(20)
    assert world.test.maybe_provide_from(provider, "1") is None

    world.provider(DummyIntProvider)
    with pytest.raises(RuntimeError):
        world.test.maybe_provide_from(world.get(DummyIntProvider), 10)



from typing import Dict, Hashable

from antidote.core import DependencyValue, Provider, Scope


class DummyProvider(Provider):

    def clone(self, keep_singletons_cache: bool) -> Provider:
        return DummyProvider(self.data, self.singleton)

    def __init__(self, data: Dict = None, singleton: bool = True):
        super().__init__()
        self.singleton = singleton
        self.data = data

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.data

    def provide(self, dependency, container):
        return DependencyValue(self.data[dependency],
                               scope=Scope.singleton() if self.singleton else None)


class DummyFactoryProvider(Provider):

    def clone(self, keep_singletons_cache: bool) -> Provider:
        return DummyFactoryProvider(self.data, self.singleton)

    def __init__(self, data: Dict = None, singleton: bool = True):
        super().__init__()
        self.singleton = singleton
        self.data = data or dict()

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.data

    def provide(self, dependency, container):
        return DependencyValue(self.data[dependency](container),
                               scope=Scope.singleton() if self.singleton else None)



from typing import Dict, Tuple, Type, cast

from .. import API

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class FinalMeta(type):
    def __new__(mcs: 'Type[FinalMeta]',
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object]
                ) -> 'FinalMeta':
        for b in bases:
            if isinstance(b, FinalMeta):
                raise TypeError(f"Type '{b.__name__}' cannot be inherited.")

        return cast(FinalMeta, super().__new__(mcs, name, bases, namespace))


@API.private
class AbstractMeta(type):
    def __new__(mcs: 'Type[AbstractMeta]',
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                abstract: bool = False
                ) -> 'AbstractMeta':
        namespace[_ABSTRACT_FLAG] = abstract
        if not abstract:
            for b in bases:
                if isinstance(b, mcs) and not getattr(b, _ABSTRACT_FLAG):
                    raise TypeError(
                        f"Cannot inherit a service which is not defined abstract. "
                        f"Consider defining {b} abstract by adding 'abstract=True' as a "
                        f"metaclass parameter. If so, Antidote won't use it.")

        return cast(AbstractMeta, super().__new__(mcs, name, bases, namespace))



import enum
from typing import Set, TypeVar

from .debug import debug_repr, short_id
from .immutable import FinalImmutable, Immutable
from .meta import AbstractMeta, FinalMeta
from .slots import SlotRecord
from .. import API

Im = TypeVar('Im', bound=Immutable)


@API.private
class Default(enum.Enum):
    sentinel = enum.auto()


@API.private
class Copy(enum.Enum):
    IDENTICAL = enum.auto()

    @staticmethod
    def immutable(current: Im, **kwargs: object) -> Im:
        return type(current)(**{
            attr: getattr(current, attr) if value is Copy.IDENTICAL else value
            for attr, value in kwargs.items()
        })


__all__ = ['debug_repr', 'short_id', 'FinalImmutable', 'Immutable', 'AbstractMeta',
           'FinalMeta', 'SlotRecord', 'API', 'Default', 'Copy']



"""
Utilities to overrides dependencies in Antidote. Those can only be used in a test world
created by :py:func:`.world.test.clone` with :code:`overridable=True`. Dependencies
can be overridden multiple times, it won't raise any error. Overridden dependencies also
won't be taken into account when Antidote checks for duplicates. Hence you may do the
following:

.. doctest:: world_override

    >>> from antidote import world
    >>> with world.test.clone():
    ...     world.test.override.singleton('test', 1)
    ...     # Override a second time a singleton
    ...     world.test.override.singleton('test', 2)
    ...     world.get[int]("test")
    2

"""
from ._override import factory, provider, singleton

__all__ = ['singleton', 'factory', 'provider']



import pytest

from antidote import Scope, world


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


def test_new():
    s = world.scopes.new(name="test")
    assert isinstance(s, Scope)
    assert s.name == "test"


@pytest.mark.parametrize('expectation,name', [
    (pytest.raises(TypeError, match=".*name.*str.*"), object()),
    (pytest.raises(ValueError, match=".*reserved.*"), Scope.singleton().name),
    (pytest.raises(ValueError, match=".*reserved.*"), Scope.sentinel().name),
    (pytest.raises(ValueError, match=".*empty.*"), "")
])
def test_invalid_new_scope_name(expectation, name):
    with expectation:
        world.scopes.new(name=name)


def test_no_duplicate_scope():
    world.scopes.new(name='dummy')
    with pytest.raises(ValueError, match=".*already exists.*"):
        world.scopes.new(name='dummy')


def test_reset():
    s = world.scopes.new(name='dummy')
    world.scopes.reset(s)  # should just work, the reset() is tested directly in providers


@pytest.mark.parametrize('expectation,scope', [
    (pytest.raises(TypeError, match=".*scope.*Scope.*"), object()),
    (pytest.raises(ValueError, match=".*Cannot reset.*"), Scope.singleton()),
    (pytest.raises(ValueError, match=".*Cannot reset.*"), Scope.sentinel())
])
def test_invalid_reset(expectation, scope):
    with expectation:
        world.scopes.reset(scope)


def test_reset_unknown_scope():
    s = world.scopes.new(name='1')

    with world.test.empty():
        with pytest.raises(ValueError, match=".*Unknown.*"):
            world.scopes.reset(s)



import inspect
from typing import Callable, Dict, Hashable, Optional

# @formatter:off
cimport cython
from cpython.ref cimport PyObject

from antidote._providers.service cimport Parameterized
from antidote.core.container cimport (DependencyResult, FastProvider, Header,
                                      HeaderObject, header_is_singleton, Scope,
                                      RawContainer, header_flag_cacheable)
from .._internal.utils import debug_repr
from ..core import DependencyDebug
from ..core.exceptions import DependencyNotFoundError
# @formatter:on

ctypedef Py_ssize_t Py_hash_t

cdef extern from "Python.h":
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    PyObject*PyDict_GetItem(PyObject *p, PyObject *key)
    PyObject*PyObject_Call(PyObject *callable, PyObject *args,
                           PyObject *kwargs) except NULL
    PyObject*PyObject_CallObject(PyObject *callable, PyObject *args) except NULL
    void Py_DECREF(PyObject *o)

cdef:
    tuple empty_tuple = tuple()


cdef class FactoryProvider(FastProvider):
    cdef:
        dict __factories

    def __cinit__(self, dict factories = None):
        self.__factories = factories or dict()  # type: Dict[FactoryDependency, Factory]

    def __repr__(self):
        return f"{type(self).__name__}(factories={self.__factories})"

    cpdef FactoryProvider clone(self, bint keep_singletons_cache):
        cdef:
            ClonedFactoryProvider provider
        provider = ClonedFactoryProvider.__new__(ClonedFactoryProvider,
                                                 self.__factories.copy())
        provider.__keep_singletons_cache = keep_singletons_cache
        return provider

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Parameterized):
            dependency = dependency.wrapped
        return (isinstance(dependency, FactoryDependency)
                and dependency in self.__factories)

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        cdef:
            Factory factory

        dependency_factory = dependency.wrapped if isinstance(dependency, Parameterized) else dependency
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory]
        except KeyError:
            return None

        dependencies = []
        wired = []
        if factory.dependency is not None:
            dependencies.append(factory.dependency)
            if isinstance(factory.dependency, type) \
                    and inspect.isclass(factory.dependency):
                wired.append(factory.dependency.__call__)
        else:
            wired.append(factory.function)

        header = HeaderObject(factory.header)
        return DependencyDebug(
            debug_repr(dependency),
            scope=header.to_scope(self._bound_container()),
            wired=wired,
            dependencies=dependencies)

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            PyObject*factory
            bint is_parameterized = PyObject_IsInstance(dependency, <PyObject*> Parameterized)
            PyObject*dependency_factory = (<PyObject*> (<Parameterized> dependency).wrapped
                                           if is_parameterized else
                                           dependency)

        if not PyObject_IsInstance(dependency_factory, <PyObject*> FactoryDependency):
            return

        factory = PyDict_GetItem(<PyObject*> self.__factories, dependency_factory)
        if factory is NULL:
            return

        if (<Factory> factory).function is None:
            (<RawContainer> container).fast_get(
                <PyObject*> (<Factory> factory).dependency,
                result)
            if result.value is NULL:
                raise DependencyNotFoundError((<Factory> factory).dependency)
            assert header_is_singleton(
                result.header), "factory dependency is expected to be a singleton"
            (<Factory> factory).function = <object> result.value
            Py_DECREF(result.value)

        if is_parameterized:
            result.header = (<Factory> factory).header
            result.value = PyObject_Call(
                <PyObject*> (<Factory> factory).function,
                <PyObject*> empty_tuple,
                <PyObject*> (<Parameterized> dependency).parameters
            )
        else:
            result.header = (<Factory> factory).header | header_flag_cacheable()
            result.value = PyObject_CallObject(
                <PyObject*> (<Factory> factory).function,
                NULL
            )

    def register(self,
                 output: type,
                 *,
                 Scope scope,
                 factory: Callable[..., object] = None,
                 factory_dependency: Hashable = None
                 ) -> FactoryDependency:
        cdef:
            Header header
            Factory f
        assert inspect.isclass(output) \
               and (factory is None or factory_dependency is None) \
               and (factory is None or callable(factory)) \
               and (isinstance(scope, Scope) or scope is None)
        with self._bound_container_ensure_not_frozen():
            dependency = FactoryDependency(output, factory or factory_dependency)
            self._bound_container_raise_if_exists(dependency)

            f = Factory.__new__(Factory)
            f.origin = <PyObject*> self
            f.header = HeaderObject.from_scope(scope).header
            if factory_dependency:
                f.dependency = factory_dependency
                f.function = None
            else:
                f.dependency = None
                f.function = factory
            self.__factories[dependency] = f
            return dependency

cdef class ClonedFactoryProvider(FactoryProvider):
    cdef:
        bint __keep_singletons_cache

    cdef object provider_type(self):
        return FactoryProvider

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            Factory f
            PyObject* factory
            bint is_parameterized = PyObject_IsInstance(dependency, <PyObject*> Parameterized)
            PyObject*dependency_factory = (<PyObject*> (<Parameterized> dependency).wrapped
                                           if is_parameterized else
                                           dependency)

        if not PyObject_IsInstance(dependency_factory, <PyObject*> FactoryDependency):
            return

        factory = PyDict_GetItem(<PyObject*> self.__factories, dependency_factory)
        if factory is NULL:
            return

        if (<Factory> factory).origin != <PyObject*> self \
                and (<Factory> factory).dependency is not None:
            f = Factory.__new__(Factory)
            f.origin = <PyObject*> self
            f.header = (<Factory> factory).header
            f.dependency = (<Factory> factory).dependency
            if self.__keep_singletons_cache:
                f.function = (<Factory> factory).function
            else:
                f.function = None
            PyDict_SetItem(<PyObject*> self.__factories, dependency_factory, <PyObject*> f)
            factory = <PyObject*> f

        if (<Factory> factory).function is None:
            (<RawContainer> container).fast_get(
                <PyObject*> (<Factory> factory).dependency,
                result)
            if result.value is NULL:
                raise DependencyNotFoundError((<Factory> factory).dependency)
            assert header_is_singleton(
                result.header), "factory dependency is expected to be a singleton"
            (<Factory> factory).function = <object> result.value
            Py_DECREF(result.value)

        if is_parameterized:
            result.header = (<Factory> factory).header
            result.value = PyObject_Call(
                <PyObject*> (<Factory> factory).function,
                <PyObject*> empty_tuple,
                <PyObject*> (<Parameterized> dependency).parameters
            )
        else:
            result.header = (<Factory> factory).header | header_flag_cacheable()
            result.value = PyObject_CallObject(
                <PyObject*> (<Factory> factory).function,
                NULL
            )


@cython.final
cdef class FactoryDependency:
    cdef:
        readonly object output
        readonly object factory
        Py_hash_t _hash

    def __init__(self, object output, object factory):
        self.output = output
        self.factory = factory
        self._hash = hash((output, factory))

    def __repr__(self) -> str:
        return f"FactoryDependency({self})"

    def __antidote_debug_repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{debug_repr(self.output)} @ {debug_repr(self.factory)}"

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        cdef:
            FactoryDependency fd

        if not isinstance(other, FactoryDependency):
            return False

        fd = <FactoryDependency> other
        return (self._hash == fd._hash
                and (self.output is fd.output
                     or self.output == fd.output)
                and (self.factory is fd.factory
                     or self.factory == fd.factory))  # noqa

@cython.freelist(64)
@cython.final
cdef class Factory:
    cdef:
        Header header
        void* origin
        object function
        object dependency

    def __repr__(self):
        return (f"{type(self).__name__}(function={self.function}, "
                f"dependency={self.dependency})")



import random
import threading
import time
from typing import Callable

from antidote import factory, world
from antidote.core import Container


class A:
    pass


class B:
    pass


class ThreadSafetyTest:
    n_threads = 10
    __state = None

    @classmethod
    def run(cls, target: Callable[[], object], n_threads=None):
        threads = [threading.Thread(target=target)
                   for _ in range(n_threads or cls.n_threads)]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

    @staticmethod
    def random_delay(a=0.01, b=None):
        b = b or a
        time.sleep(a + b * random.random())

    @staticmethod
    def unique_id():
        return threading.get_ident(), random.random()

    @classmethod
    def check_locked(cls, failures: list):
        tid = ThreadSafetyTest.unique_id()
        cls.__state = tid
        ThreadSafetyTest.random_delay()
        if cls.__state != tid:
            failures.append(1)


def delayed_new_class(cls):
    def f() -> cls:
        ThreadSafetyTest.random_delay()
        return cls()

    return f


def test_container_instantiation_safety():
    with world.test.new():
        build_a = factory(delayed_new_class(A), singleton=True)
        build_b = factory(delayed_new_class(B), singleton=False)

        singleton_got = []
        non_singleton_got = []

        def worker():
            singleton_got.append(world.get(A @ build_a))
            non_singleton_got.append(world.get(B @ build_b))

        ThreadSafetyTest.run(worker)
        assert len(set(singleton_got)) == 1
        assert len(set(non_singleton_got)) == ThreadSafetyTest.n_threads


# Be sure not have used a fixture to create a new test world as it will
# interfere with this test.
def test_world_safety():
    singletons = []

    def worker():
        with world.test.empty():
            tid = ThreadSafetyTest.unique_id()
            world.test.singleton("x", tid)
            ThreadSafetyTest.random_delay()
            singletons.append((tid, tid == world.get("x")))

    ThreadSafetyTest.run(worker)
    assert ThreadSafetyTest.n_threads == len({tid for (tid, _) in singletons})
    assert all(equal for (_, equal) in singletons)


def test_state_init_safety():
    from antidote._internal import state
    from antidote._internal import world as world_utils

    old_new_container = world_utils.new_container

    called = 0

    def new_container():
        nonlocal called
        called += 1
        ThreadSafetyTest.random_delay()
        return old_new_container()

    state.reset()
    world_utils.new_container = new_container

    try:
        ThreadSafetyTest.run(state.init)
    finally:
        world_utils.new_container = old_new_container

    assert state.current_container() is not None
    assert called == 1


def test_state_override_safety():
    from antidote._internal import state

    state.init()
    container = state.current_container()

    def create(c):
        assert c is container
        return Container()

    def worker():
        ThreadSafetyTest.random_delay()
        state.override(create)

    ThreadSafetyTest.run(worker)



import os
import pathlib

from setuptools import Extension, find_packages, setup

here = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))

with open(str(here / 'README.rst'), 'r') as f:
    readme = f.read()

ext_modules = []
install_requires = [
    'typing_extensions; python_version < "3.9.0"'
]

# Ideally this would be done with a installation flag...
if os.environ.get("ANTIDOTE_COMPILED") == "true":
    install_requires.append('fastrlock>=0.7,<0.9')

    from Cython.Build import cythonize
    from Cython.Compiler import Options

    cython_options = set(os.environ.get("ANTIDOTE_CYTHON_OPTIONS", "").split(","))
    if "all" in cython_options:
        cython_options |= {"annotate", "debug", "trace", "profile"}

    extension_extras = {}
    cythonize_extras = {}
    compiler_directives_extras = {}

    if "annotate" in cython_options:
        Options.annotate = True
        cythonize_extras['annotate'] = True

    if "debug" in cython_options:
        cythonize_extras['gdb_debug'] = True

    if "trace" in cython_options:  # line_profiler / coverage
        directive_defaults = Options.get_directive_defaults()
        directive_defaults['linetrace'] = True
        directive_defaults['binding'] = True
        extension_extras['define_macros'] = [('CYTHON_TRACE', '1')]

    if "profile" in cython_options:  # cProfile
        compiler_directives_extras["profile"] = True


    def generate_extensions():
        extensions = []
        for root, _, filenames in os.walk('src'):
            for filename in filenames:
                if filename.endswith('.pyx'):
                    path = os.path.join(root, filename)
                    module = path[4:].replace('/', '.').rsplit('.', 1)[0]
                    extensions.append(Extension(module, [path], language='c++',
                                                **extension_extras))

        return extensions


    ext_modules = cythonize(generate_extensions(),
                            compiler_directives=dict(language_level=3,
                                                     boundscheck=False,
                                                     wraparound=False,
                                                     annotation_typing=False,
                                                     cdivision=True,
                                                     **compiler_directives_extras),
                            **cythonize_extras)

setup(
    name='antidote',
    use_scm_version=dict(write_to="src/antidote/_internal/scm_version.py"),
    description='Dependency injection.',
    long_description=readme,
    author='Benjamin Rabier',
    url='https://github.com/Finistere/antidote',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={"antidote": ["py.typed"]},
    include_dirs=["src"],
    ext_modules=ext_modules,
    python_requires='>=3.6,<4',
    install_requires=install_requires,
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable ',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='dependency injection',
    zip_safe=False,
)



from typing import Dict, Hashable, Optional

import pytest

from antidote.core.container import (Container, DependencyValue, RawContainer,
                                     RawProvider, Scope)
from antidote.core.exceptions import DuplicateDependencyError
from antidote.core.utils import DependencyDebug
from antidote.exceptions import (DependencyCycleError, DependencyInstantiationError,
                                 DependencyNotFoundError, FrozenWorldError)
from .utils import DummyFactoryProvider, DummyProvider


class A:
    def __init__(self, *args):
        pass


class B:
    def __init__(self, *args):
        pass


class C:
    def __init__(self, *args):
        pass


class ServiceWithNonMetDependency:
    def __init__(self, dependency):
        pass


@pytest.fixture()
def container():
    return RawContainer()


def test_dependency_repr():
    o = object()
    d = DependencyValue(o, scope=Scope.singleton())

    assert 'singleton' in repr(d)
    assert repr(o) in repr(d)


def test_scope_repr():
    s = Scope("test")
    assert "test" in repr(s)
    assert "test" in str(s)


def test_get(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        A: lambda _: A(),
        ServiceWithNonMetDependency: lambda _: ServiceWithNonMetDependency(),
    }
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}

    assert isinstance(container.get(A), A)
    assert isinstance(container.provide(A), DependencyValue)
    assert 'Antidote' == container.get('name')
    assert 'Antidote' == container.provide('name').unwrapped

    with pytest.raises(DependencyNotFoundError):
        container.get(object)

    with pytest.raises(DependencyInstantiationError):
        container.get(ServiceWithNonMetDependency)


def test_singleton(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        A: lambda _: A(),
        B: lambda _: B(),
    }

    service = container.get(A)
    assert container.get(A) is service
    assert container.provide(A).unwrapped is service
    assert container.provide(A).scope is Scope.singleton()

    container.get(DummyFactoryProvider).singleton = False
    another_service = container.get(B)
    assert container.get(B) is not another_service
    assert container.provide(B).unwrapped is not another_service
    assert container.provide(B).scope is None

    assert container.get(A) == service


def test_dependency_cycle_error(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        A: lambda _: container.get(B),
        B: lambda _: container.get(C),
        C: lambda _: container.get(A),
    }

    with pytest.raises(DependencyCycleError):
        container.get(A)

    with pytest.raises(DependencyCycleError):
        container.get(B)

    with pytest.raises(DependencyCycleError):
        container.get(C)


def test_dependency_instantiation_error(container: RawContainer):
    container.add_provider(DummyFactoryProvider)

    def raise_error():
        raise RuntimeError()

    container.get(DummyFactoryProvider).data = {
        A: lambda _: container.get(B),
        B: lambda _: container.get(C),
        C: lambda _: raise_error(),
    }

    with pytest.raises(DependencyInstantiationError, match=".*C.*"):
        container.get(C)

    with pytest.raises(DependencyInstantiationError, match=".*A.*"):
        container.get(A)


def test_providers_property(container: RawContainer):
    x = object()
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = dict(x=x)
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = dict(y=lambda c: c.get('y'))

    assert len(container.providers) == 2
    for provider in container.providers:
        if isinstance(provider, DummyProvider):
            assert provider.data == dict(x=x)
        else:
            assert 'y' in provider.data


def test_scope_property(container: RawContainer):
    assert container.scopes == []

    s1 = container.create_scope('1')
    assert container.scopes == [s1]

    s2 = container.create_scope('2')
    assert container.scopes == [s1, s2]


def test_repr_str(container: RawContainer):
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}

    assert repr(container.get(DummyProvider)) in repr(container)
    assert str(container.get(DummyProvider)) in str(container)


def test_freeze(container: RawContainer):
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    container.freeze()

    with pytest.raises(FrozenWorldError):
        container.add_provider(DummyFactoryProvider)


def test_freezing_locked(container: RawContainer):
    with container.locked(freezing=True):
        pass

    container.freeze()

    with pytest.raises(FrozenWorldError):
        with container.locked(freezing=True):
            pass


def test_provider_property(container: RawContainer):
    container.add_provider(DummyProvider)
    assert container.providers == [container.get(DummyProvider)]


@pytest.mark.parametrize('singleton', [True, False])
def test_clone(container: RawContainer, singleton: bool):
    class A:
        pass

    container.add_provider(DummyFactoryProvider)
    provider = container.get(DummyFactoryProvider)
    provider.data = {'a': lambda c: A()}
    provider.singleton = singleton
    original = container.get('a')

    cloned = container.clone(keep_singletons=True)
    assert cloned.get(DummyFactoryProvider) is not provider
    assert cloned.get(DummyFactoryProvider).data == provider.data
    assert isinstance(cloned.get('a'), A)
    assert (cloned.get('a') is original) == singleton

    cloned = container.clone(keep_singletons=False)
    assert cloned.get(DummyFactoryProvider) is not provider
    assert cloned.get(DummyFactoryProvider).data == provider.data
    assert isinstance(cloned.get('a'), A)
    assert cloned.get('a') is not original


def test_providers_must_properly_clone(container: RawContainer):
    class DummySelf(RawProvider):
        def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
            return self

    container.add_provider(DummySelf)

    with pytest.raises(RuntimeError, match="(?i).*provider.*instance.*"):
        container.clone(keep_singletons=False,
                        keep_scopes=False)


def test_providers_must_properly_clone2(container: RawContainer):
    container.add_provider(DummyProvider)
    p = container.get(DummyProvider)

    class DummyRegistered(RawProvider):
        def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
            return p

    container.add_provider(DummyRegistered)
    with pytest.raises(RuntimeError, match="(?i).*provider.*fresh instance.*"):
        container.clone(keep_singletons=False,
                        keep_scopes=False)


@pytest.mark.filterwarnings("ignore:Debug information")
def test_raise_if_exists(container: RawContainer):
    container.raise_if_exists(object())  # Nothing should happen

    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    with pytest.raises(DuplicateDependencyError, match=".*DummyProvider.*"):
        container.raise_if_exists('name')

    class DummyProviderWithDebug(DummyProvider):
        def debug(self, dependency) -> DependencyDebug:
            return DependencyDebug("debug_info", scope=Scope.singleton())

    container.add_provider(DummyProviderWithDebug)
    container.get(DummyProviderWithDebug).data = {'hello': 'world'}
    with pytest.raises(DuplicateDependencyError,
                       match=".*DummyProviderWithDebug.*\\ndebug_info"):
        container.raise_if_exists('hello')


def test_scope(container: RawContainer):
    class ScopeProvider(RawProvider):
        dependencies: Dict[object, DependencyValue] = {}

        def exists(self, dependency):
            return dependency in self.dependencies

        def maybe_provide(self, dependency: Hashable, container: Container
                          ) -> Optional[DependencyValue]:
            try:
                return self.dependencies[dependency]
            except KeyError:
                return None

    container.add_provider(ScopeProvider)

    scope = container.create_scope('dummy')
    x = object()
    y = object()
    ScopeProvider.dependencies[1] = DependencyValue(x, scope=scope)
    assert container.get(1) is x

    ScopeProvider.dependencies[1] = DependencyValue(y, scope=scope)
    # Using cache
    assert container.get(1) is x
    assert container.provide(1) == DependencyValue(x, scope=scope)

    container.reset_scope(scope)
    assert container.get(1) is y
    assert container.provide(1) == DependencyValue(y, scope=scope)


def test_sanity_checks(container: RawContainer):
    # Cannot register twice the same kind of provider
    container.add_provider(DummyProvider)
    with pytest.raises(AssertionError):
        container.add_provider(DummyProvider)

    container.create_scope('test')
    with pytest.raises(AssertionError):
        container.create_scope('test')


def test_already_frozen(container: RawContainer):
    container.freeze()

    with pytest.raises(FrozenWorldError):
        container.freeze()


def test_debug(container: RawContainer):
    with pytest.raises(DependencyNotFoundError):
        container.debug(object())



from typing import cast, overload

from antidote import (Constants, From, Get, Provide, Service, const, factory, inject,
                      world)
from antidote._compatibility.typing import Annotated, Protocol


def test_constants_typing() -> None:
    with world.test.new():
        class MyService(Service):
            pass

        world.test.singleton('test', [])
        world.get[list]('test').append(1)
        world.lazy[list]('test').get().append(2)

        class Conf(Constants):
            A = const[list]("a")
            B = const[list]("b", default=[2])

            @inject
            def provide_const(self,
                              name: str,
                              arg: object,
                              my_service: Provide[MyService] = None
                              ) -> object:
                assert isinstance(my_service, MyService)
                return []

        Conf().A.append(1)


def test_annotated_typing() -> None:
    with world.test.new():
        class Dummy:
            def hello(self) -> 'Dummy':
                return self

        world.test.singleton('dummy', Dummy())

        @factory
        def build_dummy() -> Dummy:
            return Dummy()

        @inject
        def f(dummy: Annotated[Dummy, Get('dummy')] = None) -> Dummy:  # noqa: F821, E501
            assert dummy is not None
            return dummy

        assert f().hello() is world.get[Dummy](
            Annotated[Dummy, Get('dummy')])  # noqa: F821, E501
        assert world.get[Dummy](Annotated[Dummy, Get('dummy')]) \
               is world.get[Dummy]('dummy')

        @inject
        def g(dummy: Annotated[Dummy, From(build_dummy)] = None) -> Dummy:
            assert dummy is not None
            return dummy

        assert g().hello() is world.get[Dummy](Dummy @ build_dummy)


def test_proper_typing_assert_none() -> None:
    with world.test.new():
        class MyService(Service):
            pass

        @inject
        def f(my_service: Provide[MyService] = None) -> MyService:
            # We never expect it to be None, but it Mypy will now
            # understand that my_service may not be provided.
            assert my_service is not None
            return my_service

        s: MyService = f()  # noqa: F841

        # You can also overload the function, if you want a more accurate type definition:
        from typing import overload

        @overload
        def g(my_service: MyService) -> MyService: ...  # noqa: E704

        @overload
        def g() -> MyService: ...  # noqa: E704

        @inject
        def g(my_service: Provide[MyService] = None) -> MyService:
            assert my_service is not None
            return my_service

        s2: MyService = g()  # noqa: F841


def test_proper_typing_protocol() -> None:
    with world.test.new():
        class MyService(Service):
            pass

        @inject
        def f(my_service: Provide[MyService]) -> MyService:
            return my_service

        class FProtocol(Protocol):
            @overload
            def __call__(self, my_service: Service) -> MyService: ...  # noqa: E704

            @overload
            def __call__(self) -> MyService: ...  # noqa: E704

            def __call__(self, my_service: Service = None) -> MyService: ...  # noqa: E704

        ff = cast(FProtocol, f)

        s: MyService = ff()  # noqa: F841



import inspect
from typing import Callable, Dict, TypeVar, Union, cast

from .exceptions import DoubleInjectionError
from .injection import inject
from .wiring import Methods, Wiring
from .._internal import API
from .._internal.wrapper import is_wrapper

C = TypeVar('C', bound=type)
AnyF = Union[Callable[..., object], staticmethod, classmethod]


@API.private
def wire_class(cls: C, wiring: Wiring) -> C:
    if not (isinstance(cls, type) and inspect.isclass(cls)):
        raise TypeError(f"Expecting a class, got a {type(cls)}")

    methods: Dict[str, AnyF] = dict()
    if isinstance(wiring.methods, Methods):
        assert wiring.methods is Methods.ALL  # Sanity check
        for name, member in cls.__dict__.items():
            if (name in {'__call__', '__init__'}
                    or not (name.startswith("__") and name.endswith("__"))):
                if (inspect.isfunction(member)
                        or is_wrapper(member)  # we may need to raise DoubleInjectionError
                        or isinstance(member, (staticmethod, classmethod))):
                    methods[name] = cast(AnyF, member)
    else:
        for method_name in wiring.methods:
            try:
                attr = cls.__dict__[method_name]
            except KeyError as e:
                raise AttributeError(method_name) from e

            if not (callable(attr)
                    or isinstance(attr, (staticmethod, classmethod))):
                raise TypeError(f"{method_name} is not a (static/class) method. "
                                f"Found: {type(attr)}")
            methods[method_name] = cast(AnyF, attr)

    for name, method in methods.items():
        try:
            injected_method = inject(
                method,
                dependencies=wiring.dependencies,
                auto_provide=wiring.auto_provide,
                strict_validation=False
            )
        except DoubleInjectionError:
            if wiring.raise_on_double_injection:
                raise
        else:
            if injected_method is not method:  # If something has changed
                setattr(cls, name, injected_method)

    return cls
